@@ -59,9 +59,11 @@ typedef Eigen::Matrix<double,Eigen::Dynamic,Eigen::Dynamic,Eigen::ColMajor> Matr
 #include "g2o/config.h"
 #include "g2o/types/slam2d/types_slam2d.h"
 #include "g2o/types/slam3d/types_slam3d.h"
-#include "g2o/edge_se3_xyzprior.h"
+#include "g2o/edge_se3_xyzprior.h" // Include after types_slam3d.h to be ignored on newest g2o versions
 #include "g2o/edge_se3_gravity.h"
 #include "g2o/edge_sbacam_gravity.h"
+#include "g2o/edge_xy_prior.h"  // Include after types_slam2d.h to be ignored on newest g2o versions
+#include "g2o/edge_xyz_prior.h" // Include after types_slam3d.h to be ignored on newest g2o versions
 #ifdef G2O_HAVE_CSPARSE
 #include "g2o/solvers/csparse/linear_solver_csparse.h"
 #endif
@@ -108,6 +110,8 @@ enum {
     PARAM_OFFSET=0,
 };
 
+#define MULTICAM_OFFSET 10 // 10 means max 10 cameras per pose
+
 namespace rtabmap {
 
 bool OptimizerG2O::available()
@@ -137,6 +141,28 @@ bool OptimizerG2O::isCholmodAvailable()
 #endif
 }
 
+OptimizerG2O::OptimizerG2O(const ParametersMap & parameters) :
+		Optimizer(parameters),
+		solver_(Parameters::defaultg2oSolver()),
+		optimizer_(Parameters::defaultg2oOptimizer()),
+		pixelVariance_(Parameters::defaultg2oPixelVariance()),
+		robustKernelDelta_(Parameters::defaultg2oRobustKernelDelta()),
+		baseline_(Parameters::defaultg2oBaseline())
+{
+#ifdef RTABMAP_G2O
+	// Issue on android, have to explicitly register this type when using fixed root prior below
+	if(!g2o::Factory::instance()->knowsTag("CACHE_SE3_OFFSET"))
+	{
+#if defined(RTABMAP_G2O_CPP11) && RTABMAP_G2O_CPP11 == 1
+		g2o::Factory::instance()->registerType("CACHE_SE3_OFFSET", std::make_unique<g2o::HyperGraphElementCreator<g2o::CacheSE3Offset> >());
+#else
+		g2o::Factory::instance()->registerType("CACHE_SE3_OFFSET", new g2o::HyperGraphElementCreator<g2o::CacheSE3Offset>);
+#endif
+	}
+#endif
+	parseParameters(parameters);
+}
+
 void OptimizerG2O::parseParameters(const ParametersMap & parameters)
 {
 	Optimizer::parseParameters(parameters);
@@ -224,17 +250,17 @@ std::map<int, Transform> OptimizerG2O::optimize(
 		if(solver_ == 3)
 		{
 			//eigen
-			auto linearSolver = g2o::make_unique<SlamLinearEigenSolver>();
+			auto linearSolver = std::make_unique<SlamLinearEigenSolver>();
 			linearSolver->setBlockOrdering(false);
-			blockSolver = g2o::make_unique<SlamBlockSolver>(std::move(linearSolver));
+			blockSolver = std::make_unique<SlamBlockSolver>(std::move(linearSolver));
 		}
 #ifdef G2O_HAVE_CHOLMOD
 		else if(solver_ == 2)
 		{
 			//chmold
-			auto linearSolver = g2o::make_unique<SlamLinearCholmodSolver>();
+			auto linearSolver = std::make_unique<SlamLinearCholmodSolver>();
 			linearSolver->setBlockOrdering(false);
-			blockSolver = g2o::make_unique<SlamBlockSolver>(std::move(linearSolver));
+			blockSolver = std::make_unique<SlamBlockSolver>(std::move(linearSolver));
 		}
 #endif
 #ifdef G2O_HAVE_CSPARSE
@@ -242,16 +268,16 @@ std::map<int, Transform> OptimizerG2O::optimize(
 		{
 
 			//csparse
-			auto linearSolver = g2o::make_unique<SlamLinearCSparseSolver>();
+			auto linearSolver = std::make_unique<SlamLinearCSparseSolver>();
 			linearSolver->setBlockOrdering(false);
-			blockSolver = g2o::make_unique<SlamBlockSolver>(std::move(linearSolver));
+			blockSolver = std::make_unique<SlamBlockSolver>(std::move(linearSolver));
 		}
 #endif
 		else
 		{
 			//pcg
-			auto linearSolver = g2o::make_unique<SlamLinearPCGSolver>();
-			blockSolver = g2o::make_unique<SlamBlockSolver>(std::move(linearSolver));
+			auto linearSolver = std::make_unique<SlamLinearPCGSolver>();
+			blockSolver = std::make_unique<SlamBlockSolver>(std::move(linearSolver));
 		}
 
 		if(optimizer_ == 1)
@@ -310,14 +336,28 @@ std::map<int, Transform> OptimizerG2O::optimize(
 		}
 #endif
 		// detect if there is a global pose prior set, if so remove rootId
-		if(!priorsIgnored())
+		bool hasGravityConstraints = false;
+		if(!priorsIgnored() || (!isSlam2d() && gravitySigma() > 0))
 		{
 			for(std::multimap<int, Link>::const_iterator iter=edgeConstraints.begin(); iter!=edgeConstraints.end(); ++iter)
 			{
-				if(iter->second.from() == iter->second.to() && iter->second.type() == Link::kPosePrior)
+				if(iter->second.from() == iter->second.to())
 				{
-					rootId = 0;
-					break;
+					if(!priorsIgnored() && iter->second.type() == Link::kPosePrior)
+					{
+						rootId = 0;
+						break;
+					}
+					else if(!isSlam2d() &&
+							gravitySigma() > 0 &&
+							iter->second.type() == Link::kGravity)
+					{
+						hasGravityConstraints = true;
+						if(priorsIgnored())
+						{
+							break;
+						}
+					}
 				}
 			}
 		}
@@ -325,7 +365,7 @@ std::map<int, Transform> OptimizerG2O::optimize(
 		int landmarkVertexOffset = poses.rbegin()->first+1;
 		std::map<int, bool> isLandmarkWithRotation;
 
-		UDEBUG("fill poses to g2o...");
+		UDEBUG("fill poses to g2o... (rootId=%d hasGravityConstraints=%d isSlam2d=%d)", rootId, hasGravityConstraints?1:0, isSlam2d()?1:0);
 		for(std::map<int, Transform>::const_iterator iter = poses.begin(); iter!=poses.end(); ++iter)
 		{
 			UASSERT(!iter->second.isNull());
@@ -339,6 +379,7 @@ std::map<int, Transform> OptimizerG2O::optimize(
 					v2->setEstimate(g2o::SE2(iter->second.x(), iter->second.y(), iter->second.theta()));
 					if(id == rootId)
 					{
+						UDEBUG("Set %d fixed", id);
 						v2->setFixed(true);
 					}
 					vertex = v2;
@@ -361,6 +402,11 @@ std::map<int, Transform> OptimizerG2O::optimize(
 					{
 						g2o::VertexSE2 * v2 = new g2o::VertexSE2();
 						v2->setEstimate(g2o::SE2(iter->second.x(), iter->second.y(), iter->second.theta()));
+						if(id == rootId)
+						{
+							UDEBUG("Set %d fixed", id);
+							v2->setFixed(true);
+						}
 						vertex = v2;
 						isLandmarkWithRotation.insert(std::make_pair(id, true));
 						id = landmarkVertexOffset - id;
@@ -382,8 +428,9 @@ std::map<int, Transform> OptimizerG2O::optimize(
 					pose = a.linear();
 					pose.translation() = a.translation();
 					v3->setEstimate(pose);
-					if(id == rootId)
+					if(id == rootId && !hasGravityConstraints)
 					{
+						UDEBUG("Set %d fixed", id);
 						v3->setFixed(true);
 					}
 					vertex = v3;
@@ -412,6 +459,11 @@ std::map<int, Transform> OptimizerG2O::optimize(
 						pose = a.linear();
 						pose.translation() = a.translation();
 						v3->setEstimate(pose);
+						if(id == rootId && !hasGravityConstraints)
+						{
+							UDEBUG("Set %d fixed", id);
+							v3->setFixed(true);
+						}
 						vertex = v3;
 						isLandmarkWithRotation.insert(std::make_pair(id, true));
 						id = landmarkVertexOffset - id;
@@ -422,8 +474,50 @@ std::map<int, Transform> OptimizerG2O::optimize(
 					continue;
 				}
 			}
-			vertex->setId(id);
-			UASSERT_MSG(optimizer.addVertex(vertex), uFormat("cannot insert vertex %d!?", iter->first).c_str());
+			if(vertex == 0)
+			{
+				UERROR("Could not create vertex for node %d", id);
+			}
+			else
+			{
+				vertex->setId(id);
+				UASSERT_MSG(optimizer.addVertex(vertex), uFormat("cannot insert vertex %d!?", iter->first).c_str());
+			}
+		}
+
+		// Setup root prior (fixed x,y,z,yaw)
+		if(!isSlam2d() && rootId !=0 && hasGravityConstraints)
+		{
+			g2o::VertexSE3* v1 = dynamic_cast<g2o::VertexSE3*>(optimizer.vertex(rootId));
+			if(v1)
+			{
+				g2o::EdgeSE3Prior * e = new g2o::EdgeSE3Prior();
+				e->setVertex(0, v1);
+				Eigen::Affine3d a = poses.at(rootId).toEigen3d();
+				Eigen::Isometry3d pose;
+				pose = a.linear();
+				pose.translation() = a.translation();
+				e->setMeasurement(pose);
+				e->setParameterId(0, PARAM_OFFSET);
+				Eigen::Matrix<double, 6, 6> information = Eigen::Matrix<double, 6, 6>::Identity()*10e6;
+				// pitch and roll not fixed
+				information(3,3) = information(4,4) = 1;
+				e->setInformation(information);
+				if (!optimizer.addEdge(e))
+				{
+					delete e;
+					UERROR("Map: Failed adding fixed constraint of rootid %d, set as fixed instead", rootId);
+					v1->setFixed(true);
+				}
+				else
+				{
+					UDEBUG("Set %d fixed with prior (have gravity constraints)", rootId);
+				}
+			}
+			else
+			{
+				UERROR("Map: Failed adding fixed constraint of rootid %d (not found in added vertices)", rootId);
+			}
 		}
 
 		UDEBUG("fill edges to g2o...");
@@ -441,11 +535,37 @@ std::map<int, Transform> OptimizerG2O::optimize(
 
 			if(id1 == id2)
 			{
-				if(iter->second.type() == Link::kPosePrior && !priorsIgnored())
+				if(iter->second.type() == Link::kPosePrior && !priorsIgnored() &&
+				   (!landmarksIgnored() || id1>0))
 				{
+					int idTag= id1;
+					if(id1<0)
+					{
+						// landmark prior, offset ids
+						id1 = landmarkVertexOffset - id1;
+						id2 = landmarkVertexOffset - id2;
+					}
+
 					if(isSlam2d())
 					{
-						if (1 / static_cast<double>(iter->second.infMatrix().at<double>(5,5)) >= 9999.0)
+						if(idTag < 0 && !isLandmarkWithRotation.at(idTag))
+						{
+							g2o::EdgeXYPrior * priorEdge = new g2o::EdgeXYPrior();
+							g2o::VertexPointXY* v1 = (g2o::VertexPointXY*)optimizer.vertex(id1);
+							priorEdge->setVertex(0, v1);
+							priorEdge->setMeasurement(Eigen::Vector2d(iter->second.transform().x(), iter->second.transform().y()));
+							Eigen::Matrix<double, 2, 2> information = Eigen::Matrix<double, 2, 2>::Identity();
+							if(!isCovarianceIgnored())
+							{
+								information(0,0) = iter->second.infMatrix().at<double>(0,0); // x-x
+								information(0,1) = iter->second.infMatrix().at<double>(0,1); // x-y
+								information(1,0) = iter->second.infMatrix().at<double>(1,0); // y-x
+								information(1,1) = iter->second.infMatrix().at<double>(1,1); // y-y
+							}
+							priorEdge->setInformation(information);
+							edge = priorEdge;
+						}
+						else if (1 / static_cast<double>(iter->second.infMatrix().at<double>(5,5)) >= 9999.0)
 						{
 							g2o::EdgeSE2XYPrior * priorEdge = new g2o::EdgeSE2XYPrior();
 							g2o::VertexSE2* v1 = (g2o::VertexSE2*)optimizer.vertex(id1);
@@ -488,12 +608,36 @@ std::map<int, Transform> OptimizerG2O::optimize(
 					}
 					else
 					{
-						if (1 / static_cast<double>(iter->second.infMatrix().at<double>(3,3)) >= 9999.0 ||
+						if(idTag < 0 && !isLandmarkWithRotation.at(idTag))
+						{
+							//XYZ case
+							g2o::EdgeXYZPrior * priorEdge = new g2o::EdgeXYZPrior();
+							g2o::VertexPointXYZ* v1 = (g2o::VertexPointXYZ*)optimizer.vertex(id1);
+							priorEdge->setVertex(0, v1);
+							priorEdge->setMeasurement(Eigen::Vector3d(iter->second.transform().x(), iter->second.transform().y(), iter->second.transform().z()));
+							priorEdge->setParameterId(0, PARAM_OFFSET);
+							Eigen::Matrix<double, 3, 3> information = Eigen::Matrix<double, 3, 3>::Identity();
+							if(!isCovarianceIgnored())
+							{
+								information(0,0) = iter->second.infMatrix().at<double>(0,0); // x-x
+								information(0,1) = iter->second.infMatrix().at<double>(0,1); // x-y
+								information(0,2) = iter->second.infMatrix().at<double>(0,2); // x-z
+								information(1,0) = iter->second.infMatrix().at<double>(1,0); // y-x
+								information(1,1) = iter->second.infMatrix().at<double>(1,1); // y-y
+								information(1,2) = iter->second.infMatrix().at<double>(1,2); // y-z
+								information(2,0) = iter->second.infMatrix().at<double>(2,0); // z-x
+								information(2,1) = iter->second.infMatrix().at<double>(2,1); // z-y
+								information(2,2) = iter->second.infMatrix().at<double>(2,2); // z-z
+							}
+							priorEdge->setInformation(information);
+							edge = priorEdge;
+						}
+						else if (1 / static_cast<double>(iter->second.infMatrix().at<double>(3,3)) >= 9999.0 ||
 						    1 / static_cast<double>(iter->second.infMatrix().at<double>(4,4)) >= 9999.0 ||
 							1 / static_cast<double>(iter->second.infMatrix().at<double>(5,5)) >= 9999.0)
 						{
 							//GPS XYZ case
-							EdgeSE3XYZPrior * priorEdge = new EdgeSE3XYZPrior();
+							g2o::EdgeSE3XYZPrior * priorEdge = new g2o::EdgeSE3XYZPrior();
 							g2o::VertexSE3* v1 = (g2o::VertexSE3*)optimizer.vertex(id1);
 							priorEdge->setVertex(0, v1);
 							priorEdge->setMeasurement(Eigen::Vector3d(iter->second.transform().x(), iter->second.transform().y(), iter->second.transform().z()));
@@ -806,14 +950,17 @@ std::map<int, Transform> OptimizerG2O::optimize(
 		UDEBUG("Initial optimization...");
 		optimizer.initializeOptimization();
 
-		UASSERT_MSG(optimizer.verifyInformationMatrices(true),
-				"This error can be caused by (1) bad covariance matrix "
+		if(!optimizer.verifyInformationMatrices(true))
+		{
+			UERROR("This error can be caused by (1) bad covariance matrix "
 				"set in odometry messages "
 				"(see requirements in g2o::OptimizableGraph::verifyInformationMatrices() function) "
 				"or that (2) PCL and g2o hadn't "
 				"been built both with or without \"-march=native\" compilation "
 				"flag (if one library is built with this flag and not the other, "
 				"this is causing Eigen to not work properly, resulting in segmentation faults).");
+			return optimizedPoses;
+		}
 
 		UINFO("g2o optimizing begin (max iterations=%d, robust=%d)", iterations(), isRobust()?1:0);
 		int it = 0;
@@ -941,7 +1088,7 @@ std::map<int, Transform> OptimizerG2O::optimize(
 
 				if(i>0 && optimizer.activeRobustChi2() > 1000000000000.0)
 				{
-					UWARN("g2o: Large optimimzation error detected (%f), aborting optimization!");
+					UERROR("g2o: Large optimimzation error detected (%f), aborting optimization!");
 					return optimizedPoses;
 				}
 
@@ -984,7 +1131,7 @@ std::map<int, Transform> OptimizerG2O::optimize(
 
 		if(optimizer.activeRobustChi2() > 1000000000000.0)
 		{
-			UWARN("g2o: Large optimimzation error detected (%f), aborting optimization!");
+			UERROR("g2o: Large optimimzation error detected (%f), aborting optimization!");
 			return optimizedPoses;
 		}
 
@@ -1253,7 +1400,7 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 		int rootId,
 		const std::map<int, Transform> & poses,
 		const std::multimap<int, Link> & links,
-		const std::map<int, CameraModel> & models,
+		const std::map<int, std::vector<CameraModel> > & models,
 		std::map<int, cv::Point3f> & points3DMap,
 		const std::map<int, std::map<int, FeatureBA> > & wordReferences,
 		std::set<int> * outliers)
@@ -1267,7 +1414,7 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 	{
 		g2o::SparseOptimizer optimizer;
 		//optimizer.setVerbose(ULogger::level()==ULogger::kDebug);
-#if defined(RTABMAP_G2O_CPP11) and not defined(RTABMAP_ORB_SLAM)
+#if defined(RTABMAP_G2O_CPP11) && !defined(RTABMAP_ORB_SLAM)
 		std::unique_ptr<g2o::BlockSolver_6_3::LinearSolverType> linearSolver;
 #else
 		g2o::BlockSolver_6_3::LinearSolverType * linearSolver = 0;
@@ -1280,7 +1427,7 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 		{
 			//eigen
 #ifdef RTABMAP_G2O_CPP11
-			linearSolver = g2o::make_unique<g2o::LinearSolverEigen<g2o::BlockSolver_6_3::PoseMatrixType> >();
+			linearSolver = std::make_unique<g2o::LinearSolverEigen<g2o::BlockSolver_6_3::PoseMatrixType> >();
 #else
 			linearSolver = new g2o::LinearSolverEigen<g2o::BlockSolver_6_3::PoseMatrixType>();
 #endif
@@ -1290,7 +1437,7 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 		{
 			//chmold
 #ifdef RTABMAP_G2O_CPP11
-			linearSolver = g2o::make_unique<g2o::LinearSolverCholmod<g2o::BlockSolver_6_3::PoseMatrixType> >();
+			linearSolver = std::make_unique<g2o::LinearSolverCholmod<g2o::BlockSolver_6_3::PoseMatrixType> >();
 #else
 			linearSolver = new g2o::LinearSolverCholmod<g2o::BlockSolver_6_3::PoseMatrixType>();
 #endif
@@ -1301,7 +1448,7 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 		{
 			//csparse
 #ifdef RTABMAP_G2O_CPP11
-			linearSolver = g2o::make_unique<g2o::LinearSolverCSparse<g2o::BlockSolver_6_3::PoseMatrixType> >();
+			linearSolver = std::make_unique<g2o::LinearSolverCSparse<g2o::BlockSolver_6_3::PoseMatrixType> >();
 #else
 			linearSolver = new g2o::LinearSolverCSparse<g2o::BlockSolver_6_3::PoseMatrixType>();
 #endif
@@ -1311,7 +1458,7 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 		{
 			//pcg
 #ifdef RTABMAP_G2O_CPP11
-			linearSolver = g2o::make_unique<g2o::LinearSolverPCG<g2o::BlockSolver_6_3::PoseMatrixType> >();
+			linearSolver = std::make_unique<g2o::LinearSolverPCG<g2o::BlockSolver_6_3::PoseMatrixType> >();
 #else
 			linearSolver = new g2o::LinearSolverPCG<g2o::BlockSolver_6_3::PoseMatrixType>();
 #endif
@@ -1323,7 +1470,7 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 		{
 #ifdef RTABMAP_G2O_CPP11
 			optimizer.setAlgorithm(new g2o::OptimizationAlgorithmGaussNewton(
-					g2o::make_unique<g2o::BlockSolver_6_3>(std::move(linearSolver))));
+					std::make_unique<g2o::BlockSolver_6_3>(std::move(linearSolver))));
 #else
 			optimizer.setAlgorithm(new g2o::OptimizationAlgorithmGaussNewton(new g2o::BlockSolver_6_3(linearSolver)));
 #endif
@@ -1331,9 +1478,9 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 		else
 #endif
 		{
-#if defined(RTABMAP_G2O_CPP11) and not defined(RTABMAP_ORB_SLAM)
+#if defined(RTABMAP_G2O_CPP11) && !defined(RTABMAP_ORB_SLAM)
 			optimizer.setAlgorithm(new g2o::OptimizationAlgorithmLevenberg(
-					g2o::make_unique<g2o::BlockSolver_6_3>(std::move(linearSolver))));
+					std::make_unique<g2o::BlockSolver_6_3>(std::move(linearSolver))));
 #else
 			optimizer.setAlgorithm(new g2o::OptimizationAlgorithmLevenberg(new g2o::BlockSolver_6_3(linearSolver)));
 #endif
@@ -1346,50 +1493,55 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 			if(iter->first > 0)
 			{
 				// Get camera model
-				std::map<int, CameraModel>::const_iterator iterModel = models.find(iter->first);
-				UASSERT(iterModel != models.end() && iterModel->second.isValidForProjection());
+				std::map<int, std::vector<CameraModel> >::const_iterator iterModel = models.find(iter->first);
+				UASSERT(iterModel != models.end() && !iterModel->second.empty());
+				for(size_t i=0; i<iterModel->second.size(); ++i)
+				{
+					UASSERT(iterModel->second[i].isValidForProjection());
 
-				Transform camPose = iter->second * iterModel->second.localTransform();
+					Transform camPose = iter->second * iterModel->second[i].localTransform();
 
-				// Add node's pose
-				UASSERT(!camPose.isNull());
+					// Add node's pose
+					UASSERT(!camPose.isNull());
 #ifdef RTABMAP_ORB_SLAM
-				g2o::VertexSE3Expmap * vCam = new g2o::VertexSE3Expmap();
+					g2o::VertexSE3Expmap * vCam = new g2o::VertexSE3Expmap();
 #else
-				g2o::VertexCam * vCam = new g2o::VertexCam();
+					g2o::VertexCam * vCam = new g2o::VertexCam();
 #endif
 
-				Eigen::Affine3d a = camPose.toEigen3d();
+					Eigen::Affine3d a = camPose.toEigen3d();
 #ifdef RTABMAP_ORB_SLAM
-				a = a.inverse();
-				vCam->setEstimate(g2o::SE3Quat(a.linear(), a.translation()));
+					a = a.inverse();
+					vCam->setEstimate(g2o::SE3Quat(a.linear(), a.translation()));
 #else
-				g2o::SBACam cam(Eigen::Quaterniond(a.linear()), a.translation());
-				cam.setKcam(
-						iterModel->second.fx(),
-						iterModel->second.fy(),
-						iterModel->second.cx(),
-						iterModel->second.cy(),
-						iterModel->second.Tx()<0.0?-iterModel->second.Tx()/iterModel->second.fx():baseline_); // baseline in meters
-				vCam->setEstimate(cam);
+					g2o::SBACam cam(Eigen::Quaterniond(a.linear()), a.translation());
+					cam.setKcam(
+							iterModel->second[i].fx(),
+							iterModel->second[i].fy(),
+							iterModel->second[i].cx(),
+							iterModel->second[i].cy(),
+							iterModel->second[i].Tx()<0.0?-iterModel->second[i].Tx()/iterModel->second[i].fx():baseline_); // baseline in meters
+					vCam->setEstimate(cam);
 #endif
-				vCam->setId(iter->first);
+					vCam->setId(iter->first*MULTICAM_OFFSET + i);
 
-				// negative root means that all other poses should be fixed instead of the root
-				vCam->setFixed((rootId >= 0 && iter->first == rootId) || (rootId < 0 && iter->first != -rootId));
+					// negative root means that all other poses should be fixed instead of the root
+					vCam->setFixed((rootId >= 0 && iter->first == rootId) || (rootId < 0 && iter->first != -rootId));
 
-				/*UDEBUG("cam %d (fixed=%d) fx=%f fy=%f cx=%f cy=%f Tx=%f baseline=%f t=%s",
-						iter->first,
-						vCam->fixed()?1:0,
-						iterModel->second.fx(),
-						iterModel->second.fy(),
-						iterModel->second.cx(),
-						iterModel->second.cy(),
-						iterModel->second.Tx(),
-						iterModel->second.Tx()<0.0?-iterModel->second.Tx()/iterModel->second.fx():baseline_,
-						camPose.prettyPrint().c_str());*/
+					/*UDEBUG("camPose %d (camid=%d) (fixed=%d) fx=%f fy=%f cx=%f cy=%f Tx=%f baseline=%f t=%s",
+							iter->first,
+							vCam->id(),
+							vCam->fixed()?1:0,
+							iterModel->second[i].fx(),
+							iterModel->second[i].fy(),
+							iterModel->second[i].cx(),
+							iterModel->second[i].cy(),
+							iterModel->second[i].Tx(),
+							iterModel->second[i].Tx()<0.0?-iterModel->second[i].Tx()/iterModel->second[i].fx():baseline_,
+							camPose.prettyPrint().c_str());*/
 
-				UASSERT_MSG(optimizer.addVertex(vCam), uFormat("cannot insert vertex %d!?", iter->first).c_str());
+					UASSERT_MSG(optimizer.addVertex(vCam), uFormat("cannot insert cam vertex %d (pose=%d)!?", vCam->id(), iter->first).c_str());
+				}
 			}
 		}
 
@@ -1421,11 +1573,12 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 
 						Eigen::MatrixXd information = Eigen::MatrixXd::Identity(3, 3) * 1.0/(gravitySigma()*gravitySigma());
 
-						g2o::VertexCam* v1 = (g2o::VertexCam*)optimizer.vertex(id1);
+						g2o::VertexCam* v1 = (g2o::VertexCam*)optimizer.vertex(id1*MULTICAM_OFFSET);
 						EdgeSBACamGravity* priorEdge(new EdgeSBACamGravity());
-						std::map<int, CameraModel>::const_iterator iterModel = models.find(iter->first);
-						UASSERT(iterModel != models.end() && !iterModel->second.localTransform().isNull());
-						priorEdge->setCameraInvLocalTransform(iterModel->second.localTransform().inverse().toEigen3d().linear());
+						std::map<int, std::vector<CameraModel> >::const_iterator iterModel = models.find(iter->first);
+						// Gravity constraint added only to first camera of a pose
+						UASSERT(iterModel != models.end() && !iterModel->second.empty() && !iterModel->second[0].localTransform().isNull());
+						priorEdge->setCameraInvLocalTransform(iterModel->second[0].localTransform().inverse().toEigen3d().linear());
 						priorEdge->setMeasurement(m);
 						priorEdge->setInformation(information);
 						priorEdge->vertices()[0] = v1;
@@ -1450,15 +1603,17 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 					}
 
 					// between cameras, not base_link
-					Transform camLink = models.at(id1).localTransform().inverse()*iter->second.transform()*models.at(id2).localTransform();
-					//UDEBUG("added edge %d->%d (in cam frame=%s)",
-					//		id1,
-					//		id2,
-					//		camLink.prettyPrint().c_str());
+					Transform camLink = models.at(id1)[0].localTransform().inverse()*iter->second.transform()*models.at(id2)[0].localTransform();
+					/*UDEBUG("added edge %d->%d (camIDs %d->%d) (in cam frame=%s)",
+							id1,
+							id2,
+							id1*MULTICAM_OFFSET,
+							id2*MULTICAM_OFFSET,
+							camLink.prettyPrint().c_str());*/
 #ifdef RTABMAP_ORB_SLAM
 					EdgeSE3Expmap * e = new EdgeSE3Expmap();
-					g2o::VertexSE3Expmap* v1 = (g2o::VertexSE3Expmap*)optimizer.vertex(id1);
-					g2o::VertexSE3Expmap* v2 = (g2o::VertexSE3Expmap*)optimizer.vertex(id2);
+					g2o::VertexSE3Expmap* v1 = (g2o::VertexSE3Expmap*)optimizer.vertex(id1*MULTICAM_OFFSET);
+					g2o::VertexSE3Expmap* v2 = (g2o::VertexSE3Expmap*)optimizer.vertex(id2*MULTICAM_OFFSET);
 
 					Transform camPose1 = Transform::fromEigen3d(v1->estimate()).inverse();
 					Transform camPose2Inv = Transform::fromEigen3d(v2->estimate());
@@ -1466,8 +1621,8 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 					camLink =  camPose1 * camPose1 * camLink * camPose2Inv * camPose2Inv;
 #else
 					g2o::EdgeSBACam * e = new g2o::EdgeSBACam();
-					g2o::VertexCam* v1 = (g2o::VertexCam*)optimizer.vertex(id1);
-					g2o::VertexCam* v2 = (g2o::VertexCam*)optimizer.vertex(id2);
+					g2o::VertexCam* v1 = (g2o::VertexCam*)optimizer.vertex(id1*MULTICAM_OFFSET);
+					g2o::VertexCam* v2 = (g2o::VertexCam*)optimizer.vertex(id2*MULTICAM_OFFSET);
 #endif
 					UASSERT(v1 != 0);
 					UASSERT(v2 != 0);
@@ -1487,8 +1642,60 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 			}
 		}
 
+		UDEBUG("fill hard edges between camera 0 and other cameras (multicam)...");
+		for(std::map<int, std::vector<CameraModel> >::const_iterator iter=models.begin(); iter!=models.end(); ++iter)
+		{
+			int id = iter->first;
+			if(uContains(poses, id))
+			{
+				for(size_t i=1; i<iter->second.size(); ++i)
+				{
+					// add edge
+					// Set large information matrix to keep these links fixed
+					Eigen::Matrix<double, 6, 6> information = Eigen::Matrix<double, 6, 6>::Identity()*9999999;
+
+					// between cameras, not base_link
+					Transform camLink = iter->second[0].localTransform().inverse()*iter->second[i].localTransform();
+#ifdef RTABMAP_ORB_SLAM
+					EdgeSE3Expmap * e = new EdgeSE3Expmap();
+					g2o::VertexSE3Expmap* v1 = (g2o::VertexSE3Expmap*)optimizer.vertex(id*MULTICAM_OFFSET);
+					g2o::VertexSE3Expmap* v2 = (g2o::VertexSE3Expmap*)optimizer.vertex(id*MULTICAM_OFFSET+i);
+
+					Transform camPose1 = Transform::fromEigen3d(v1->estimate()).inverse();
+					Transform camPose2Inv = Transform::fromEigen3d(v2->estimate());
+
+					camLink =  camPose1 * camPose1 * camLink * camPose2Inv * camPose2Inv;
+#else
+					g2o::EdgeSBACam * e = new g2o::EdgeSBACam();
+					g2o::VertexCam* v1 = (g2o::VertexCam*)optimizer.vertex(id*MULTICAM_OFFSET);
+					g2o::VertexCam* v2 = (g2o::VertexCam*)optimizer.vertex(id*MULTICAM_OFFSET+i);
+#endif
+					/*UDEBUG("added edge between subcam 0->%d id:%d->%d (in cam frame=%s)",
+							i,
+							v1->id(),
+							v2->id(),
+							camLink.prettyPrint().c_str());*/
+
+					UASSERT(v1 != 0);
+					UASSERT(v2 != 0);
+					e->setVertex(0, v1);
+					e->setVertex(1, v2);
+					Eigen::Affine3d a = camLink.toEigen3d();
+					e->setMeasurement(g2o::SE3Quat(a.linear(), a.translation()));
+					e->setInformation(information);
+
+					if (!optimizer.addEdge(e))
+					{
+						delete e;
+						UERROR("Map: Failed adding constraint between %d and %d, skipping", v1->id(), v2->id());
+						return optimizedPoses;
+					}
+				}
+			}
+		}
+
 		UDEBUG("fill 3D points to g2o...");
-		const int stepVertexId = poses.rbegin()->first+1;
+		const int stepVertexId = poses.rbegin()->first*MULTICAM_OFFSET+MULTICAM_OFFSET;
 		int negVertexOffset = stepVertexId;
 		if(wordReferences.size() && wordReferences.rbegin()->first>0)
 		{
@@ -1527,22 +1734,24 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 				// set observations
 				for(std::map<int, FeatureBA>::const_iterator jter=iter->second.begin(); jter!=iter->second.end(); ++jter)
 				{
-					int camId = jter->first;
-					if(poses.find(camId) != poses.end() && optimizer.vertex(camId) != 0)
+					int poseId = jter->first;
+					int camIndex = jter->second.cameraIndex;
+					int camId = poseId*MULTICAM_OFFSET+camIndex;
+					if(poses.find(poseId) != poses.end() && optimizer.vertex(camId) != 0)
 					{
 						const FeatureBA & pt = jter->second;
 						double depth = pt.depth;
 
-						//UDEBUG("Added observation pt=%d to cam=%d (%d,%d) depth=%f", vpt3d->id()-stepVertexId, camId, (int)pt.kpt.pt.x, (int)pt.kpt.pt.y, depth);
+						//UDEBUG("Added observation pt=%d to cam=%d (%d,%d) depth=%f (camIndex=%d)", vpt3d->id()-stepVertexId, camId, (int)pt.kpt.pt.x, (int)pt.kpt.pt.y, depth, camIndex);
 
 						g2o::OptimizableGraph::Edge * e;
 						double baseline = 0.0;
 #ifdef RTABMAP_ORB_SLAM
 						g2o::VertexSE3Expmap* vcam = dynamic_cast<g2o::VertexSE3Expmap*>(optimizer.vertex(camId));
-						std::map<int, CameraModel>::const_iterator iterModel = models.find(camId);
+						std::map<int, std::vector<CameraModel> >::const_iterator iterModel = models.find(poseId);
 
-						UASSERT(iterModel != models.end() && iterModel->second.isValidForProjection());
-						baseline = iterModel->second.Tx()<0.0?-iterModel->second.Tx()/iterModel->second.fx():baseline_;
+						UASSERT(iterModel != models.end() && camIndex<iterModel->second.size() && iterModel->second[camIndex].isValidForProjection());
+						baseline = iterModel->second[camIndex].Tx()<0.0?-iterModel->second[camIndex].Tx()/iterModel->second[camIndex].fx():baseline_;
 #else
 						g2o::VertexCam* vcam = dynamic_cast<g2o::VertexCam*>(optimizer.vertex(camId));
 						baseline = vcam->estimate().baseline;
@@ -1553,15 +1762,15 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 							// stereo edge
 #ifdef RTABMAP_ORB_SLAM
 							g2o::EdgeStereoSE3ProjectXYZ* es = new g2o::EdgeStereoSE3ProjectXYZ();
-							float disparity = baseline * iterModel->second.fx() / depth;
+							float disparity = baseline * iterModel->second[camIndex].fx() / depth;
 							Eigen::Vector3d obs( pt.kpt.pt.x, pt.kpt.pt.y, pt.kpt.pt.x-disparity);
 							es->setMeasurement(obs);
 							//variance *= log(exp(1)+disparity);
 							es->setInformation(Eigen::Matrix3d::Identity() / variance);
-							es->fx = iterModel->second.fx();
-							es->fy = iterModel->second.fy();
-							es->cx = iterModel->second.cx();
-							es->cy = iterModel->second.cy();
+							es->fx = iterModel->second[camIndex].fx();
+							es->fy = iterModel->second[camIndex].fy();
+							es->cx = iterModel->second[camIndex].cx();
+							es->cy = iterModel->second[camIndex].cy();
 							es->bf = baseline*es->fx;
 							e = es;
 #else
@@ -1589,10 +1798,10 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 							Eigen::Vector2d obs( pt.kpt.pt.x, pt.kpt.pt.y);
 							em->setMeasurement(obs);
 							em->setInformation(Eigen::Matrix2d::Identity() / variance);
-							em->fx = iterModel->second.fx();
-							em->fy = iterModel->second.fy();
-							em->cx = iterModel->second.cx();
-							em->cy = iterModel->second.cy();
+							em->fx = iterModel->second[camIndex].fx();
+							em->fy = iterModel->second[camIndex].fy();
+							em->cx = iterModel->second[camIndex].cx();
+							em->cy = iterModel->second[camIndex].cy();
 							e = em;
 
 #else
@@ -1667,13 +1876,13 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 						{
 							d = ((g2o::EdgeStereoSE3ProjectXYZ*)(*iter))->measurement()[0]-((g2o::EdgeStereoSE3ProjectXYZ*)(*iter))->measurement()[2];
 						}
-						UDEBUG("Ignoring edge (%d<->%d) d=%f var=%f kernel=%f chi2=%f", (*iter)->vertex(0)->id()-stepVertexId, (*iter)->vertex(1)->id(), d, 1.0/((g2o::EdgeStereoSE3ProjectXYZ*)(*iter))->information()(0,0), (*iter)->robustKernel()->delta(), (*iter)->chi2());
+						//UDEBUG("Ignoring edge (%d<->%d) d=%f var=%f kernel=%f chi2=%f", (*iter)->vertex(0)->id()-stepVertexId, (*iter)->vertex(1)->id(), d, 1.0/((g2o::EdgeStereoSE3ProjectXYZ*)(*iter))->information()(0,0), (*iter)->robustKernel()->delta(), (*iter)->chi2());
 #else
 						if(dynamic_cast<g2o::EdgeProjectP2SC*>(*iter) != 0)
 						{
 							d = ((g2o::EdgeProjectP2SC*)(*iter))->measurement()[0]-((g2o::EdgeProjectP2SC*)(*iter))->measurement()[2];
 						}
-						UDEBUG("Ignoring edge (%d<->%d) d=%f var=%f kernel=%f chi2=%f", (*iter)->vertex(0)->id()-stepVertexId, (*iter)->vertex(1)->id(), d, 1.0/((g2o::EdgeProjectP2SC*)(*iter))->information()(0,0), (*iter)->robustKernel()->delta(), (*iter)->chi2());
+						//UDEBUG("Ignoring edge (%d<->%d) d=%f var=%f kernel=%f chi2=%f", (*iter)->vertex(0)->id()-stepVertexId, (*iter)->vertex(1)->id(), d, 1.0/((g2o::EdgeProjectP2SC*)(*iter))->information()(0,0), (*iter)->robustKernel()->delta(), (*iter)->chi2());
 #endif
 
 						cv::Point3f pt3d;
@@ -1716,10 +1925,11 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 		{
 			if(iter->first > 0)
 			{
+				int camId = iter->first*MULTICAM_OFFSET;
 #ifdef RTABMAP_ORB_SLAM
-				const g2o::VertexSE3Expmap* v = (const g2o::VertexSE3Expmap*)optimizer.vertex(iter->first);
+				const g2o::VertexSE3Expmap* v = (const g2o::VertexSE3Expmap*)optimizer.vertex(camId);
 #else
-				const g2o::VertexCam* v = (const g2o::VertexCam*)optimizer.vertex(iter->first);
+				const g2o::VertexCam* v = (const g2o::VertexCam*)optimizer.vertex(camId);
 #endif
 				if(v)
 				{
@@ -1730,7 +1940,7 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 #endif
 
 					// remove model local transform
-					t *= models.at(iter->first).localTransform().inverse();
+					t *= models.at(iter->first)[0].localTransform().inverse();
 
 					//UDEBUG("%d from=%s to=%s", iter->first, iter->second.prettyPrint().c_str(), t.prettyPrint().c_str());
 					if(t.isNull())
@@ -1754,7 +1964,7 @@ std::map<int, Transform> OptimizerG2O::optimizeBA(
 				}
 				else
 				{
-					UERROR("Vertex (pose) %d not found!?", iter->first);
+					UERROR("Vertex (pose) %d (cam=%d) not found!?", iter->first, camId);
 				}
 			}
 		}
